package cbr;

import java.text.NumberFormat;
import java.text.ParseException;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Collections;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;

import domain.SpeedLimitCalculation;
import domain.VisibilityEstimation;

public class Caze {

	public static final Random RANDOM = new Random();

	private final LocalDateTime timestamp;
	private final Map<String, Double> values;
	private final Map<String, Double> cache;
	private final Decision decision;

	public Caze(LocalDateTime timestamp, Map<String, Double> values, Decision decision) {
		this.timestamp = timestamp;
		this.values = Collections.unmodifiableMap(new HashMap<>(values));
		this.cache = new HashMap<>();
		this.decision = decision;
	}

	public LocalDateTime getTimestamp() {
		return timestamp;
	}

	public Double getAttributeValue(String attribute) {
		if (values.containsKey(attribute)) {
			return values.get(attribute);
		} else if (cache.containsKey(attribute)) {
			return cache.get(attribute);
		} else if (attribute.startsWith("Visibility")) {
			String location = attribute.substring("Visibility".length());
			Double precipitation = getAttributeValue("Precipitation" + location);
			Double temperature = getAttributeValue("Temperature" + location);
			Double windSpeed = getAttributeValue("Wind speed" + location);
			Double visibility = null;
			if (precipitation != null && temperature != null && windSpeed != null) {
				visibility = VisibilityEstimation.visibility(precipitation, temperature, windSpeed);
			}
			cache.put(attribute, visibility);
			return visibility;
		} else if (attribute.startsWith("Speed limit")) {
			String location = attribute.substring("Speed limit".length());
			Double windSpeed = getAttributeValue("Wind speed" + location);
			Double friction = getAttributeValue("Friction" + location);
			Double visibility = getAttributeValue("Visibility" + location);
			Double speedLimit = null;
			if (windSpeed != null && friction != null && visibility != null) {
				speedLimit = SpeedLimitCalculation.speedLimit(windSpeed, friction, visibility);
			}
			cache.put(attribute, speedLimit);
			return speedLimit;
		} else {
			throw new RuntimeException("Unknown attribute: " + attribute);
		}
	}

	public Decision getDecision() {
		return decision;
	}

	public Caze applySensorUncertainty() {
		Map<String, Double> newValues = new HashMap<>();
		for (Entry<String, Double> entry : values.entrySet()) {
			String attribute = entry.getKey();
			Double value = entry.getValue();
			if (attribute.startsWith("Wind speed")) {
				Double newValue = null;
				if (value != null) {
					if (value < 5.0) {
						newValue = value + (RANDOM.nextDouble() - 0.5);
						newValue = Math.max(newValue, 0.0);
					} else {
						newValue = value * (0.2 * RANDOM.nextDouble() + 0.9);
						newValue = Math.max(newValue, 0.0);
					}
				}
				newValues.put(attribute, newValue);
			} else if (attribute.startsWith("Wind direction")) {
				Double newValue = null;
				if (value != null) {
					newValue = value + (10.0 * RANDOM.nextDouble() - 5.0);
					if (newValue < 0.0) {
						newValue += 360.0;
					}
					if (newValue >= 360.0) {
						newValue -= 360.0;
					}
				}
				newValues.put(attribute, newValue);
			} else if (attribute.startsWith("Friction")) {
				newValues.put(attribute, value);
			} else if (attribute.startsWith("Temperature")) {
				Double newValue = null;
				if (value != null) {
					if (value >= -10.0 && value <= 10.0) {
						newValue = value + (0.2 * RANDOM.nextDouble() - 0.1);
					} else {
						newValue = value + (0.4 * RANDOM.nextDouble() - 0.2);
					}
				}
				newValues.put(attribute, newValue);
			} else if (attribute.startsWith("Precipitation")) {
				Double newValue = null;
				if (value != null) {
					newValue = value + (0.2 * RANDOM.nextDouble() - 0.1);
					newValue = Math.max(newValue, 0.0);
				}
				newValues.put(attribute, newValue);
			} else {
				newValues.put(attribute, value);
			}
		}
		return new Caze(timestamp, newValues, decision);
	}

	public static Caze fromCazeFileRecord(Map<String, String> record) throws ParseException {
		LocalDateTime timestamp = null;
		Map<String, Double> values = new HashMap<>();
		Decision decision = null;
		for (Entry<String, String> entry : record.entrySet()) {
			if (entry.getKey().equals("Time")) {
				timestamp = LocalDateTime.parse(entry.getValue(), DateTimeFormatter.ofPattern("uuuu-MM-dd HH:mm"));
			} else if (entry.getKey().equals("Decision")) {
				decision = Decision.valueOf(entry.getValue());
			} else {
				if (entry.getValue() != null && !entry.getValue().isEmpty()) {
					values.put(entry.getKey(), NumberFormat.getNumberInstance(Locale.forLanguageTag("nb"))
							.parse(entry.getValue()).doubleValue());
				} else {
					values.put(entry.getKey(), null);
				}
			}
		}
		if (timestamp == null) {
			throw new RuntimeException("No timestamp found");
		}
		if (decision == null) {
			throw new RuntimeException("No decision found");
		}
		return new Caze(timestamp, values, decision);
	}
}
