package ea_gui;

import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;
import java.text.ParseException;
import java.util.HashSet;
import java.util.Set;

import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;

import cbr.Caze;
import cbr.CazeFileReader;
import ea.AdultSelector;
import ea.ByteArrayIndividual;
import ea.CBRTester;
import ea.EA;
import ea.EASimilarityFunction;
import ea.FitnessProportionateReproducer;
import ea.FitnessTester;
import ea.Individual;
import ea.RankReproducer;
import ea.Reproducer;
import ea.SigmaScalingReproducer;
import ea.TournamentReproducer;
import ea.UniformReproducer;

public class MainPanel extends JPanel {

	private static final long serialVersionUID = 1L;

	private ReproducerPanel reproducerPanel = new ReproducerPanel();
	private AdultSelectorPanel adultSelectorPanel = new AdultSelectorPanel();
	private JButton btnStart = new JButton("Start");
	private JButton btnStop = new JButton("Stop");

	private Thread eaThread;
	private boolean stopEA;

	public MainPanel() {
		super(new GridBagLayout());
		GridBagConstraints c = new GridBagConstraints();
		c.insets.bottom = 0;
		c.gridy = 0;
		add(new JLabel("Reproducer (Parent Selector) parameters"), c);
		c.insets.bottom = 50;
		c.gridy = 1;
		add(reproducerPanel, c);
		c.insets.bottom = 0;
		c.gridy = 2;
		add(new JLabel(
				"Adult Selector parameters - In replacement mode #Adults = #Children gives full replacement and #Adults < #Children gives over production"),
				c);
		c.insets.bottom = 50;
		c.gridy = 3;
		add(adultSelectorPanel, c);
		c.insets.bottom = 10;
		c.gridy = 4;
		add(btnStart, c);
		btnStart.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				start();
			}
		});
		c.gridy = 5;
		add(btnStop, c);
		btnStop.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				stop();
			}
		});
	}

	private void start() {
		Set<Individual> startIndividuals = createStartIndividuals();
		Reproducer reproducer = createReproducer();
		FitnessTester tester;
		try {
			tester = createFitnessTester();
		} catch (IOException | ParseException e) {
			e.printStackTrace();
			return;
		}
		AdultSelector selector = createAdultSelector();
		Runnable eaRunner = new Runnable() {
			@Override
			public void run() {
				stopEA = false;
				EA ea = new EA(startIndividuals, reproducer, tester, selector);
				while (ea.maxFitness() < 1.0 && !stopEA) {
					ea.nextGeneration();
				}
			}
		};
		eaThread = new Thread(eaRunner);
		eaThread.start();
	}

	private void stop() {
		stopEA = true;
	}

	private Set<Individual> createStartIndividuals() {
		Set<Individual> startIndividuals = new HashSet<Individual>();
		int noAdults = adultSelectorPanel.getAdults();
		int byteArrayLength = EASimilarityFunction.nofRequiredBytes();
		while (startIndividuals.size() < noAdults) {
			startIndividuals.add(new ByteArrayIndividual(byteArrayLength));
		}
		return startIndividuals;
	}

	private Reproducer createReproducer() {
		Reproducer reproducer = null;
		int type = reproducerPanel.getReproducer();
		int noChildren = reproducerPanel.getChildren();
		double crossover = reproducerPanel.getCrossover();
		double mutation = reproducerPanel.getMutation();
		switch (type) {
		case ReproducerPanel.FITNESS_PROPORTIONATE: {
			reproducer = new FitnessProportionateReproducer(noChildren, crossover, mutation);
			break;
		}
		case ReproducerPanel.SIGMA_SCALING: {
			reproducer = new SigmaScalingReproducer(noChildren, crossover, mutation);
			break;
		}
		case ReproducerPanel.TOURNAMENT: {
			int groupSize = reproducerPanel.getGroupSize();
			double randomProbability = reproducerPanel.getRandomProbability();
			reproducer = new TournamentReproducer(noChildren, crossover, mutation, groupSize, randomProbability);
			break;
		}
		case ReproducerPanel.RANK: {
			double maxMinRatio = reproducerPanel.getMaxMinRatio();
			reproducer = new RankReproducer(noChildren, crossover, mutation, maxMinRatio);
			break;
		}
		case ReproducerPanel.UNIFORM: {
			reproducer = new UniformReproducer(noChildren, crossover, mutation);
		}
		}
		return reproducer;
	}

	private FitnessTester createFitnessTester() throws IOException, ParseException {
		Set<Caze> cazes = CazeFileReader.readFile(new File("trainingset.csv"));
		CBRTester tester = new CBRTester(cazes);
		return tester;
	}

	private AdultSelector createAdultSelector() {
		AdultSelector selector = null;
		int type = adultSelectorPanel.getSelector();
		int noAdults = adultSelectorPanel.getAdults();
		switch (type) {
		case AdultSelectorPanel.REPLACEMENT:
			selector = new AdultSelector(noAdults, AdultSelector.REPLACEMENT_MODE);
			break;
		case AdultSelectorPanel.MIXING:
			selector = new AdultSelector(noAdults, AdultSelector.MIXING_MODE);
			break;
		}
		return selector;
	}

}
