package model1;

import java.util.HashMap;
import java.util.Map;

import cbr.Caze;
import cbr.Decision;
import cbr.SimilarityFunction;

public class Model1SimilarityFunction implements SimilarityFunction {

	public static final String[] ATTRIBUTES = { "Wind speed Fokstugu", "Wind direction Fokstugu", "Friction Fokstugu",
			"Temperature Fokstugu", "Wind speed Avsjoen", "Wind direction Avsjoen", "Friction Avsjoen",
			"Temperature Avsjoen", "Wind speed Hjerkinn", "Wind direction Hjerkinn", "Temperature Hjerkinn" };

	private static final Map<String, Double> ROAD_DIRECTIONS;

	static {
		ROAD_DIRECTIONS = new HashMap<>();
		ROAD_DIRECTIONS.put("Wind direction Avsjoen", 60.0);
		ROAD_DIRECTIONS.put("Wind direction Fokstugu", 45.0);
		ROAD_DIRECTIONS.put("Wind direction Hjerkinn", 0.0);
	}

	private Map<String, Double> weights = new HashMap<>();

	public Model1SimilarityFunction() {
		weights.put("Wind speed Fokstugu", 0.001);
		weights.put("Wind direction Fokstugu", 0.050);
		weights.put("Friction Fokstugu", 0.122);
		weights.put("Temperature Fokstugu", 0.277);
		weights.put("Wind speed Avsjoen", 0.001);
		weights.put("Wind direction Avsjoen", 0.002);
		weights.put("Friction Avsjoen", 0.001);
		weights.put("Temperature Avsjoen", 0.233);
		weights.put("Wind speed Hjerkinn", 0.016);
		weights.put("Wind direction Hjerkinn", 0.018);
		weights.put("Temperature Hjerkinn", 0.279);
	}

	@Override
	public double calculateSimilarity(Caze caze, Caze query) {
		double similarity = 0;
		String[] attributes = ATTRIBUTES;
		for (String attribute : attributes) {
			double weight = weights.get(attribute);
			Double cazeValue = caze.getAttributeValue(attribute);
			Decision cazeDecision = caze.getDecision();
			Double queryValue = query.getAttributeValue(attribute);
			if (attribute.startsWith("Wind speed")) {
				similarity += weight * windSpeedSimilarity(cazeValue, cazeDecision, queryValue);
			} else if (attribute.startsWith("Wind direction")) {
				double roadDirection = ROAD_DIRECTIONS.get(attribute);
				similarity += weight * windDirectionSimilarity(cazeValue, cazeDecision, queryValue, roadDirection);
			} else if (attribute.startsWith("Friction")) {
				similarity += weight * frictionSimilarity(cazeValue, cazeDecision, queryValue);
			} else if (attribute.startsWith("Visibility")) {
				similarity += weight * visibilitySimilarity(cazeValue, cazeDecision, queryValue);
			} else if (attribute.startsWith("Temperature")) {
				similarity += weight * temperatureSimilarity(cazeValue, queryValue);
			} else if (attribute.startsWith("Speed limit")) {
				similarity += weight * speedLimitSimilarity(cazeValue, cazeDecision, queryValue);
			} else {
				throw new RuntimeException("Unknown attribute: " + attribute);
			}
		}
		return similarity;
	}

	private double windSpeedSimilarity(Double cazeValue, Decision cazeDecision, Double queryValue) {
		if (cazeValue == null || queryValue == null) {
			return 0.0;
		}
		double fc = 1.0 / (1.0 + Math.exp(-(cazeValue - 12.5) / 5.0));
		double fq = 1.0 / (1.0 + Math.exp(-(queryValue - 12.5) / 5.0));
		double similarity;
		switch (cazeDecision) {
		case OPEN:
			similarity = 1.0 - 2.0 * (fq - fc);
			break;
		case CLOSED:
			similarity = 1.0 - 2.0 * (fc - fq);
			break;
		default:
			similarity = 1.0 - 2.0 * Math.abs(fq - fc);
			break;
		}
		return Math.max(0.0, Math.min(1.0, similarity));
	}

	private double windDirectionSimilarity(Double cazeValue, Decision cazeDecision, Double queryValue,
			double roadDirection) {
		if (cazeValue == null || queryValue == null) {
			return 0.0;
		}
		double fc = roadRelativeAngle(cazeValue, roadDirection);
		double fq = roadRelativeAngle(queryValue, roadDirection);
		double similarity;
		switch (cazeDecision) {
		case OPEN:
			similarity = 1.0 - (fq - fc) / 90.0;
			break;
		case CLOSED:
			similarity = 1.0 - (fc - fq) / 90.0;
			break;
		default:
			similarity = 1.0 - Math.abs(fq - fc) / 90.0;
			break;
		}
		return Math.max(0.0, Math.min(1.0, similarity));
	}

	private double frictionSimilarity(Double cazeValue, Decision cazeDecision, Double queryValue) {
		if (cazeValue == null || queryValue == null) {
			return 0.0;
		}
		double fc = Math.log(cazeValue);
		double fq = Math.log(queryValue);
		double similarity;
		switch (cazeDecision) {
		case OPEN:
			similarity = 1.0 - (fc - fq) / Math.log(4.0);
			break;
		case CLOSED:
			similarity = 1.0 - (fq - fc) / Math.log(4.0);
			break;
		default:
			similarity = 1.0 - Math.abs(fq - fc) / Math.log(4.0);
			break;
		}
		return Math.max(0.0, Math.min(1.0, similarity));
	}

	private double visibilitySimilarity(Double cazeValue, Decision cazeDecision, Double queryValue) {
		if (cazeValue == null || queryValue == null) {
			return 0.0;
		}
		double fc = Math.log(cazeValue);
		double fq = Math.log(queryValue);
		double similarity;
		switch (cazeDecision) {
		case OPEN:
			similarity = 1.0 - (fc - fq) / Math.log(4.0);
			break;
		case CLOSED:
			similarity = 1.0 - (fq - fc) / Math.log(4.0);
			break;
		default:
			similarity = 1.0 - Math.abs(fq - fc) / Math.log(4.0);
			break;
		}
		return Math.max(0.0, Math.min(1.0, similarity));
	}

	private double temperatureSimilarity(Double cazeValue, Double queryValue) {
		if (cazeValue == null || queryValue == null) {
			return 0.0;
		}
		double fc = 1.0 / (1.0 + Math.exp(-cazeValue / 2.0));
		double fq = 1.0 / (1.0 + Math.exp(-queryValue / 2.0));
		double similarity = 1.0 - 2.0 * Math.abs(fq - fc);
		return Math.max(0.0, Math.min(1.0, similarity));
	}

	private double speedLimitSimilarity(Double cazeValue, Decision cazeDecision, Double queryValue) {
		if (cazeValue == null || queryValue == null) {
			return 0.0;
		}
		double fc = cazeValue;
		double fq = queryValue;
		double similarity;
		switch (cazeDecision) {
		case OPEN:
			similarity = 1.0 - (fc - fq) / 40.0;
			break;
		case CLOSED:
			similarity = 1.0 - (fq - fc) / 40.0;
			break;
		default:
			similarity = 1.0 - Math.abs(fq - fc) / 40.0;
			break;
		}
		return Math.max(0.0, Math.min(1.0, similarity));
	}

	private double roadRelativeAngle(double direction, double roadDirection) {
		double angle = direction - roadDirection;
		angle = ((angle % 180.0) + 180.0) % 180.0;
		angle = Math.min(angle, 180.0 - angle);
		return angle;
	}
}
