package tests;

import java.io.File;
import java.io.IOException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import cbr.CBR;
import cbr.Caze;
import cbr.CazeFileReader;
import cbr.Decision;
import cbr.MontecarloSimulation;
import domain.SpeedLimitCalculation;
import model1.Model1SimilarityFunction;
import model2.Model2SimilarityFunction;

public class Tests {

	private static final int NOF_QUERY_SETS = 1;

	public static void main(String[] args) throws IOException, ParseException {
		System.out.println("Tests running...");
		Set<Caze> trainingSet = CazeFileReader.readFile(new File("trainingset.csv"));
		Set<Caze> testSet = CazeFileReader.readFile(new File("testset.csv"));
		CBR cbr1 = new CBR(new Model1SimilarityFunction());
		CBR cbr2 = new CBR(new Model2SimilarityFunction());
		double cbr1Performance = 0.0;
		double cbr2Performance = 0.0;
		double cbr1MontecarloPerformance = 0.0;
		double cbr2MontecarloPerformance = 0.0;
		double app1Performance = 0.0;
		double app2Performance = 0.0;
		for (int i = 0; i < NOF_QUERY_SETS; i++) {
			Set<Caze> cazeBase = new HashSet<>();
			Set<Caze> queries = new HashSet<>();
			cazeBase.addAll(trainingSet);
			while (!convoyQueryExists(queries)) {
				List<Caze> testList = new ArrayList<>(testSet);
				Collections.shuffle(testList);
				cazeBase.addAll(testList.subList(0, 192));
				queries.addAll(testList.subList(192, 384));
			}
			cbr1Performance += testCBR(cbr1, cazeBase, queries);
			cbr2Performance += testCBR(cbr2, cazeBase, queries);
			app1Performance += testMobileApplicationFormula(queries, false);
			app2Performance += testMobileApplicationFormula(queries, true);
			cbr1MontecarloPerformance += testCBRWithMontecarloSimulation(cbr1, cazeBase, queries);
			cbr2MontecarloPerformance += testCBRWithMontecarloSimulation(cbr2, cazeBase, queries);
			System.out.println((i + 1) + " of " + NOF_QUERY_SETS + " tests finished");
		}
		cbr1Performance /= NOF_QUERY_SETS;
		cbr2Performance /= NOF_QUERY_SETS;
		cbr1MontecarloPerformance /= NOF_QUERY_SETS;
		cbr2MontecarloPerformance /= NOF_QUERY_SETS;
		app1Performance /= NOF_QUERY_SETS;
		app2Performance /= NOF_QUERY_SETS;
		System.out.println("Results:");
		System.out.println("CBR-model 1: " + cbr1Performance);
		System.out.println("CBR-model 2: " + cbr2Performance);
		System.out.println("CBR-model 1 with sensor uncertainty handling: " + cbr1MontecarloPerformance);
		System.out.println("CBR-model 2 with sensor uncertainty handling: " + cbr2MontecarloPerformance);
		System.out.println("Application formula: " + app1Performance);
		System.out.println("Application formula ignoring missing attribute queries: " + app2Performance);
	}

	private static boolean convoyQueryExists(Set<Caze> queries) {
		for (Caze query : queries) {
			if (query.getDecision() == Decision.CONVOY)
				return true;
		}
		return false;
	}

	private static double testCBR(CBR cbr, Set<Caze> cazeBase, Set<Caze> queries) {
		cbr.setCazeBase(cazeBase);
		int correctOpen = 0;
		int totalOpen = 0;
		int correctConvoy = 0;
		int totalConvoy = 0;
		int correctClosed = 0;
		int totalClosed = 0;
		for (Caze query : queries) {
			Decision decision = cbr.makeDecision(query);
			switch (query.getDecision()) {
			case OPEN:
				if (decision == Decision.OPEN) {
					correctOpen++;
				}
				totalOpen++;
				break;
			case CONVOY:
				if (decision == Decision.CONVOY) {
					correctConvoy++;
				}
				totalConvoy++;
				break;
			case CLOSED:
				if (decision == Decision.CLOSED) {
					correctClosed++;
				}
				totalClosed++;
				break;
			}
		}
		int nofDecisionTypes = 0;
		double result = 0.0;
		if (totalOpen != 0) {
			nofDecisionTypes++;
			result += (double) correctOpen / (double) totalOpen;
		}
		if (totalConvoy != 0) {
			nofDecisionTypes++;
			result += (double) correctConvoy / (double) totalConvoy;
		}
		if (totalClosed != 0) {
			nofDecisionTypes++;
			result += (double) correctClosed / (double) totalClosed;
		}
		if (nofDecisionTypes == 0) {
			throw new RuntimeException("No test cases");
		}
		result /= nofDecisionTypes;
		return result;
	}

	private static double testCBRWithMontecarloSimulation(CBR cbr, Set<Caze> cazeBase, Set<Caze> queries) {
		double correctOpen = 0;
		int totalOpen = 0;
		double correctConvoy = 0;
		int totalConvoy = 0;
		double correctClosed = 0;
		int totalClosed = 0;
		for (Caze query : queries) {
			Map<Decision, Double> decisionProbabilities = MontecarloSimulation.getDecisionProbabilities(cbr, cazeBase,
					query);
			switch (query.getDecision()) {
			case OPEN:
				correctOpen += decisionProbabilities.get(Decision.OPEN);
				totalOpen++;
				break;
			case CONVOY:
				correctConvoy += decisionProbabilities.get(Decision.CONVOY);
				totalConvoy++;
				break;
			case CLOSED:
				correctClosed += decisionProbabilities.get(Decision.CLOSED);
				totalClosed++;
				break;
			}
		}
		int nofDecisionTypes = 0;
		double result = 0.0;
		if (totalOpen != 0) {
			nofDecisionTypes++;
			result += correctOpen / (double) totalOpen;
		}
		if (totalConvoy != 0) {
			nofDecisionTypes++;
			result += correctConvoy / (double) totalConvoy;
		}
		if (totalClosed != 0) {
			nofDecisionTypes++;
			result += correctClosed / (double) totalClosed;
		}
		if (nofDecisionTypes == 0) {
			throw new RuntimeException("No test cases");
		}
		result /= nofDecisionTypes;
		return result;
	}

	private static double testMobileApplicationFormula(Set<Caze> queries, boolean ignoreMissingAttributeQueries) {
		int correctOpen = 0;
		int totalOpen = 0;
		int correctConvoy = 0;
		int totalConvoy = 0;
		int correctClosed = 0;
		int totalClosed = 0;
		for (Caze query : queries) {
			Double windSpeedAvsjoen = query.getAttributeValue("Wind speed Avsjoen");
			Double frictionAvsjoen = query.getAttributeValue("Friction Avsjoen");
			Double visibilityAvsjoen = query.getAttributeValue("Visibility Avsjoen");
			Decision decisionAvsjoen = null;
			if (windSpeedAvsjoen != null && frictionAvsjoen != null && visibilityAvsjoen != null) {
				decisionAvsjoen = SpeedLimitCalculation.decision(windSpeedAvsjoen, frictionAvsjoen, visibilityAvsjoen);
			}
			Double windSpeedFokstugu = query.getAttributeValue("Wind speed Fokstugu");
			Double frictionFokstugu = query.getAttributeValue("Friction Fokstugu");
			Double visibilityFokstugu = query.getAttributeValue("Visibility Fokstugu");
			Decision decisionFokstugu = null;
			if (windSpeedFokstugu != null && frictionFokstugu != null && visibilityFokstugu != null) {
				decisionFokstugu = SpeedLimitCalculation.decision(windSpeedFokstugu, frictionFokstugu,
						visibilityFokstugu);
			}
			Decision decision = null;
			if (decisionAvsjoen == Decision.CLOSED || decisionFokstugu == Decision.CLOSED) {
				decision = Decision.CLOSED;
			} else if (decisionAvsjoen == Decision.CONVOY || decisionFokstugu == Decision.CONVOY) {
				decision = Decision.CONVOY;
			} else if (decisionAvsjoen == Decision.OPEN || decisionFokstugu == Decision.OPEN) {
				decision = Decision.OPEN;
			}
			if (decision == null && ignoreMissingAttributeQueries) {
				continue;
			}
			switch (query.getDecision()) {
			case OPEN:
				if (decision == Decision.OPEN) {
					correctOpen++;
				}
				totalOpen++;
				break;
			case CONVOY:
				if (decision == Decision.CONVOY) {
					correctConvoy++;
				}
				totalConvoy++;
				break;
			case CLOSED:
				if (decision == Decision.CLOSED) {
					correctClosed++;
				}
				totalClosed++;
				break;
			}
		}
		int nofDecisionTypes = 0;
		double result = 0.0;
		if (totalOpen != 0) {
			nofDecisionTypes++;
			result += (double) correctOpen / (double) totalOpen;
		}
		if (totalConvoy != 0) {
			nofDecisionTypes++;
			result += (double) correctConvoy / (double) totalConvoy;
		}
		if (totalClosed != 0) {
			nofDecisionTypes++;
			result += (double) correctClosed / (double) totalClosed;
		}
		if (nofDecisionTypes == 0) {
			throw new RuntimeException("No test cases");
		}
		result /= nofDecisionTypes;
		return result;
	}
}
